/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define RCI_INTERNAL
#ifndef __RCI__
#include "rci.h"
#endif

static char *kRciCookie = "RCI0";

int RciReadHeader(RciState *rci)
{
	int  err;
	RciUns8 cookie[4];
	RciUns32 hdrSize, extraSize;

	err = RciRead(rci->fd,cookie,kRciCookieLen);
	if( err )
		return err;
	
	if( RciMemCmp( kRciCookie, cookie, kRciCookieLen ) )
		return kRciNotFmtErr;
	
	// Read in header fields, then skip appropriate amount
	// according to dataStart.
	err = RciRead32BE(rci->fd,&rci->dataStart);
	if( err ) return err;
	hdrSize = rci->dataStart;
	if( hdrSize < kRciMinHdrLen )
		return kRciDataErr; // corrupt file

	// figure out how much data we need to skip after the header fields that we understand.
	extraSize = hdrSize - kRciHdrLen;
	err = RciRead16BE(rci->fd,&rci->version);
	if( err ) return err;
	// ever if the version is higher than we know about we will still try to read file

	err = RciRead16BE(rci->fd,&rci->planes);
	if( err ) return err;

	err = RciRead8(rci->fd,&rci->colorMode);
	if( err ) return err;

	// we only support RGB and grayscale in this version.
	if( rci->colorMode != kRciRGB && rci->colorMode != kRciGrayscale )
		return kRciUnsupportedMode;
	
	err = RciRead8(rci->fd,&rci->planeDepth);
	if( err ) return err;
	// we only support 8 bit planes, we just specify it so future versions can handle others.
	if( rci->planeDepth != 8 )
		return kRciDepthUnsupported;

	err = RciRead8(rci->fd,&rci->reserved);
	if( err ) return err;
	err = RciRead8(rci->fd,&rci->flags);
	if( err ) return err;

	// if it has any flags we don't recognize then forget it.
	if( rci->flags & ~(kZigZagRowsFlag) )
		return kRciUnknownFlag;

	err = RciRead32BE(rci->fd,&rci->width);
	if( err ) return err;
	err = RciRead32BE(rci->fd,&rci->height);
	if( err ) return err;

	err = RciRead32BE(rci->fd,&rci->tileWidth);
	if( err ) return err;
	err = RciRead32BE(rci->fd,&rci->tileHeight);
	if( err ) return err;

	// now skip the extra data by reading it in and throwing it away.
	// we don't seek because we only have read/write I/O routines to minimize
	// the effort of porting this library.
	if( extraSize )
	{
		RciUns8 *skip = RciAlloc(extraSize);
		if( !skip )
			return kRciNoMem;
		err = RciRead(rci->fd,skip,extraSize);
		RciFree((void *)skip);
		if( err ) return err;
	}

	#ifdef RCI_DEBUG
	printf("Read header\n");
	#endif

	return kRciSuccess;
}

int RciWriteHeader(RciState *rci)
{
	// this overrides things like dataStart and version regardless of rci's contents
	
	int err;;

	// Make sure that contents of rci aren't stupid.
	if( rci->colorMode != kRciRGB && rci->colorMode != kRciGrayscale )
		return kRciUnsupportedMode;
	if( rci->planeDepth != 8 )
		return kRciDepthUnsupported;
	if( rci->flags & ~(kZigZagRowsFlag) )
		return kRciUnknownFlag;

	// Some things we force to the only allowed values.
	rci->dataStart = kRciHdrLen;
	rci->version = kRciVersion;
	rci->reserved = 0;

	err = RciWrite(rci->fd,(const RciUns8 *)kRciCookie,kRciCookieLen);
	if( err )
		return err;


	
	err = RciWrite32BE(rci->fd,rci->dataStart);
	if( err ) return err;

	err = RciWrite16BE(rci->fd,rci->version);
	if( err ) return err;

	err = RciWrite16BE(rci->fd,rci->planes);
	if( err ) return err;

	err = RciWrite8(rci->fd,rci->colorMode);
	if( err ) return err;

	err = RciWrite8(rci->fd,rci->planeDepth);
	if( err ) return err;
	
	err = RciWrite8(rci->fd,rci->reserved);
	if( err ) return err;
	
	err = RciWrite8(rci->fd,rci->flags);
	if( err ) return err;
	
	err = RciWrite32BE(rci->fd,rci->width);
	if( err ) return err;
	err = RciWrite32BE(rci->fd,rci->height);
	if( err ) return err;
	
	err = RciWrite32BE(rci->fd,rci->tileWidth);
	if( err ) return err;
	err = RciWrite32BE(rci->fd,rci->tileHeight);
	if( err ) return err;
	
	#ifdef RCI_DEBUG
	printf("Wrote header\n");
	#endif

	return kRciSuccess;
}

int RciReadImageData(RciState *rci)
{
	RciUns32 planeSize, rowSize, dataSize, tileSize;
	RciUns32 width, height, top, left;
	RciUns16 plane;
	RciUns8 *buf = NULL, *comp = NULL;
	RciUns32 bytesUsed;
	RciUns32 numTiles;
	RciUns32 bufSize, compSize;
	RciInt32 decoSize;
	int err = 0;

	width = rci->width;
	height = rci->height;

	//calculate the size of the image data.
	planeSize = width;
	planeSize *= rci->planeDepth;
	planeSize /= 8;
	rowSize = planeSize * rci->planes;
	dataSize = rowSize * height;
	
	
	// allocate the RAM for the tile buffer
	tileSize = (rci->planeDepth>>3)*rci->tileWidth*rci->tileHeight;
	//printf("tile size is %u\n",tileSize);

	numTiles = RciNumTiles(rci);
	rci->numTiles = numTiles;

	// allocate enough space for the decompressed tiles.
	bufSize = rci->planes * (kRciTileHdrLen+tileSize) * numTiles;
	buf = (RciUns8 *)RciAlloc( bufSize );
	if( !buf )
	{
		err = kRciNoMem;
		goto cleanup;
	}

	// allocate the RAM to hold it
	rci->image = RciAlloc( dataSize );
	if( !rci->image )
	{
		err = kRciNoMem;
		goto cleanup;
	}

	#ifdef RCI_DEBUG
	printf("Starting read\n");
	#endif

	// read in the data and decompress it
	err = RciRead32BE(rci->fd,&compSize);
	if( err )
		goto cleanup;
	rci->compSize = compSize;

	// allocate enough memory for the compressed data.
	comp = (RciUns8 *)RciAlloc(compSize);
	if( !comp )
	{
		err = kRciNoMem;
		goto cleanup;
	}
	err = RciRead(rci->fd,comp,compSize);
	if( err )
		goto cleanup;
	
	// decompress data into buf.
	decoSize = RciDecompress(comp,compSize,buf,bufSize);

	if( decoSize < 0 )
	{
		err = kRciDataErr;
		goto cleanup;
	}

	rci->tile = buf;

	top = 0;
	while( top < rci->height )
	{
		left = 0;
		while( left < rci->width )
		{
			#ifdef RCI_DEBUG
			printf("Tile...");
			#endif
			for(plane = 0;plane<rci->planes;plane++)
			{
				// read a tile.
				err = RciReadPlanarTile(rci,plane,left,top,&bytesUsed);
				if( err )
					goto cleanup;
				rci->tile += bytesUsed;
#ifdef RCI_DEBUG 
				//printf(", %u",bytesUsed);
#endif
			}
			left += rci->tileWidth;
#ifdef RCI_DEBUG
			printf(" %u ", (RciUns32)(rci->tile - buf));
#endif

			#ifdef RCI_DEBUG
			printf("Done\n");
			#endif
		}
		top += rci->tileHeight;
	}
	
	#ifdef RCI_DEBUG
	printf("Done reading\n");
	#endif

cleanup:
	
	if( buf )
		RciFree( (void *)buf);
	if( comp )
		RciFree( (void *)comp);
	rci->tile = NULL;

	return err;
}


int RciWriteImageData(RciState *rci)
{
	RciUns32 dataSize,planeSize,rowSize, width, height, top, left,tileSize;
	RciUns16 plane;
	int err = 0;
	RciUns32 tileUsed;
	RciUns8 *buf = NULL, *comp = NULL;
	RciUns32 numTiles;
	RciUns32 maxSize, totalSize;
	RciInt32 compSize;

	//calculate the size of the image data.

	width = rci->width;
	height = rci->height;
	planeSize = width;
	planeSize *= rci->planeDepth;
	planeSize /= 8;
	rowSize = planeSize * rci->planes;
	dataSize = rowSize * height;
	
	// setup the stats for the filters
	RciMemSet( (void *)rci->filterStat, 0, sizeof(long) * kRciNumFilters );

	numTiles = RciNumTiles(rci);
	rci->numTiles = numTiles;

	tileSize = (rci->planeDepth>>3)*rci->tileWidth*rci->tileHeight;

	// allocate enough space for the tiles after filtering
	buf = (RciUns8 *)RciAlloc( rci->planes * (kRciTileHdrLen+tileSize) * numTiles);
	if( !buf )
	{
		err = kRciNoMem;
		goto cleanup;
	}

	
	rci->tile = buf;

	#ifdef RCI_DEBUG
	printf("Starting write\n");
	#endif

	top = 0;
	while( top < rci->height )
	{
		left = 0;
		while( left < rci->width )
		{
			#ifdef RCI_DEBUG
			printf("Tile...");
			#endif
			for(plane = 0;plane<rci->planes;plane++)
			{
				// write a tile.
				err = RciWritePlanarTile(rci,plane,left,top,&tileUsed);
				if( err )
					goto cleanup;
				// update tile pointer.
				rci->tile += tileUsed;
#ifdef RCI_DEBUG 
			//	printf(", %u",tileUsed);
#endif
			}
			left += rci->tileWidth;
#ifdef RCI_DEBUG
			printf(" %u ", (RciUns32)(rci->tile - buf));
#endif
			#ifdef RCI_DEBUG
			printf("Done\n");
			#endif
		}
		top += rci->tileHeight;
	}

	totalSize = (RciUns32)(rci->tile - buf);

	maxSize = rci->planes * (kRciTileHdrLen+tileSize) * numTiles + RCI_COMPRESS_MAX_INCREASE;
	comp = (RciUns8 *)RciAlloc( maxSize );
	if( !comp )
	{
		err = kRciNoMem;
		goto cleanup;
	}

	// now compress all of the data.
	compSize = RciCompress(buf,totalSize,comp,maxSize);
	
	if( compSize < 0 )
	{
		err = kRciUnknownErr;
		goto cleanup;
	}

	rci->compSize = compSize;

	err = RciWrite32BE(rci->fd,compSize);
	if( err )
		goto cleanup;
	err = RciWrite(rci->fd,comp,compSize);

	#ifdef RCI_DEBUG
	printf("Done writing\n");
	#endif

cleanup:
	if( buf )
		RciFree((void *)buf);
	if( comp )
		RciFree((void *)comp);
	rci->tile = NULL;

	return err;
}

int RciReadImage(RciState *rci)
{
	int err;

	err = RciReadHeader(rci);
	if(err)
		return err;
	
	err = RciReadImageData(rci);
	if( err )
		return err;
	
	return 0;
}

int RciWriteImage(RciState *rci)
{
	int err;

	err = RciWriteHeader(rci);
	if( err )
		return err;
	
	err = RciWriteImageData(rci);
	if( err )
		return err;

	return 0;
}







// this writes a planar tile given its left,top and the tile width and height
// (it will clip the tile if need be)
int RciWritePlanarTile(RciState *rci, RciUns16 plane, RciUns32 left, RciUns32 top, RciUns32 *outBytes)
{
	RciUns32 rowBytes, rows,rowStore, cols, cc;
	RciUns16 planeBytes, pixelSize;
	RciUns8 *cp,*rowp,*tile;
	RciUns32 tileSize;
	RciUns8 filterType = kRciNoFilter;
	RciUns8 filterParam = 0;

	planeBytes = rci->planeDepth >> 3;
	pixelSize = rci->planes * planeBytes;

	rowBytes = pixelSize * rci->width;
	cp = (RciUns8 *)rci->image + top * rowBytes;
	cp += pixelSize * left;
	cp += planeBytes * plane;

	
	cols = left + rci->tileWidth;
	if( cols > rci->width )
		cols = rci->width;
	cols -= left;

	rows = top + rci->tileHeight;
	if( rows > rci->height )
		rows = rci->height;
	rows -= top;
	rowStore = rows;

	rowp = cp;

	// copy the tile into the scratch buffer
//	tileSize = planeBytes * rci->tileWidth * rci->tileHeight;
	tileSize = planeBytes * rows * cols;

	tile = rci->tile;
	tile += kRciTileHdrLen; // skip the space for the filter byte and param

	RciMemSet(tile,0,tileSize);

	if( planeBytes == 1 )
	{
		while(rows--)
		{
			cp = rowp;
			
			cc = cols;
			while(cc--)
			{
				*tile++ = *cp;
				cp += pixelSize;
			}
		
			rowp += rowBytes;
		}
	}
	else if( planeBytes == 2 )
	{
		while(rows--)
		{
			cp = rowp;
			
			cc = cols;
			while(cc--)
			{
				*tile++ = cp[0];
				*tile++ = cp[1];
				cp += pixelSize;
			}
			rowp += rowBytes;
		}

	}
	
	*outBytes = (RciUns32)(tile - rci->tile);

	// do the filtering
	if( rci->filteringEnabled )
		filterType = RciFilterTile( rci->filter, rci->tile+kRciTileHdrLen, 
			rci->planeDepth,rowStore,cols,&filterParam);
	else
		filterType = kRciNoFilter;

	rci->filterStat[filterType]++;

	// this reverses the odd rows in a tile
	tile = rci->tile;
	*tile++ = filterType; // put the filter type in.
	*tile++ = filterParam; // put the fitler param in.

	if( rci->flags & kZigZagRowsFlag )
		RciZigZagTile(tile, rci->planeDepth, rowStore, cols );
	return 0;
}

// this reads a planar tile given its left,top and the tile width and height
// (it will handled the clipped ones on right and bottom automatically)
int RciReadPlanarTile(RciState *rci, RciUns16 plane, RciUns32 left, RciUns32 top, RciUns32 *outBytes)
{
	RciUns32 rowBytes, rows, cols, cc;
	RciUns16 planeBytes, pixelSize;
	RciUns8 *cp,*rowp,*tile;
	RciUns32 tileSize;
	RciUns8 filterType;
	RciUns8 filterParam = 0;

	planeBytes = rci->planeDepth >> 3;
	pixelSize = rci->planes * planeBytes;

	rowBytes = pixelSize * rci->width;
	cp = (RciUns8 *)rci->image + top * rowBytes;
	cp += pixelSize * left;
	cp += planeBytes * plane;

	
	cols = left + rci->tileWidth;
	if( cols > rci->width )
		cols = rci->width;
	cols -= left;

	rows = top + rci->tileHeight;
	if( rows > rci->height )
		rows = rci->height;
	rows -= top;

	rowp = cp;


	// read the tile into the scratch buffer
	tile = rci->tile;
	tileSize = planeBytes * rows * cols;

	filterType = *tile++; // first byte is the filter type
	filterParam = *tile++;

	// this reverses the odd rows in a tile
	if( rci->flags & kZigZagRowsFlag )
		RciZigZagTile(tile, rci->planeDepth, rows, cols );
	
	// undo the filtering.
	RciInvFilterTile(tile,rci->planeDepth, rows, cols, filterType,filterParam );

	// copy the tile from the scratch buffer
	if( planeBytes == 1 )
	{
		while(rows--)
		{
			cp = rowp;
			
			cc = cols;
			while(cc--)
			{
				*cp = *tile++;
				cp += pixelSize;
			}
		
			rowp += rowBytes;
		}
	}
	else if( planeBytes == 2 )
	{
		while(rows--)
		{
			cp = rowp;
			
			cc = cols;
			while(cc--)
			{
				cp[0] = *tile++;
				cp[1] = *tile++;
				cp += pixelSize;
			}
			rowp += rowBytes;
		}

	}
	*outBytes = (RciUns32)(tile - rci->tile);
	return 0;
}




