/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __RCI__
#define __RCI__



#ifdef __cplusplus
extern "C" {
#endif

// Define this to turn on some debugging printf's
#undef RCI_DEBUG

// NOTE: test with odd number of planes like 1,2, and 4 and make sure still works.

#include <stdlib.h> // for malloc, free
#include <string.h> // for memset, memcmp, memcpy

#ifdef RCI_DEBUG
#include <stdio.h> // for printf
#endif

// Types
typedef unsigned char RciUns8;
typedef unsigned short RciUns16;
typedef unsigned long RciUns32;

typedef short RciInt16;
typedef long RciInt32;

typedef unsigned char BxBool;

enum {
	kRciVersion = 0,
	kRciHdrLen = 32,
	kRciMinHdrLen = 32, // min size required should later versions add fields.
	kRciCookieLen = 4,
	kRciTileHdrLen = 2
};



enum {
	kRciDefaultTileWidth = 32,
	kRciDefaultTileHeight = 16
};


enum {
	kRciRGB = 0,
	kRciGrayscale,
};

enum {
	kRciNoFilter = 0,
	kRciSubFilter = 1, // difference from left pixel
	kRciPaethFilter = 2, // weird linear function
	kRciUpFilter = 3, // difference from above pixel
	kRciAvgFilter = 4, // difference from average of left and above pixels
	kRciOffsetFilter = 5, // offset from average or median type value(currently median)
	kRciLocoFilter = 6, // the filter from Loco/JPEG-LS
	kRciNumFilters
};

// if the huerististic isn't lower than this
// for any of the filters, then we don't filter the tile.
#define kRciFilterCutoff 0.2

enum {
	kRciSuccess = 0, // no error
	kRciEOFErr = -1, // unexpected EOF
	kRciNotFmtErr = -2, // not in the expected format
	kRciUnsupportedErr = -3, // file uses an unsupported option
	kRciUnknownErr = -4, // we don't know what happened, but it fucked up
	kRciDataErr = -5, // the data appears to be corrupt or invaliD
	kRciStateErr = -6, // generally means programmer screwup in parameters,calling order...
	kRciNoMem = -7, // an allocation failed.
	kRciDiskFull = -8, // not enough space for write operation
	kRciReadErr = -9, // error during read operation
	kRciWriteErr = -10, // error during write operation
	kRciDepthUnsupported = -11, // plane depth in file or requested is not supported
	kRciUnsupportedMode = -12, // color mode specified is not supported.
	kRciUnknownFlag = -13 // unknown bit specified in flags field 
};

enum {
	kRciNativeFmt,
	kRciRRGBFmt,
	kRciRawFmt
};

// these are the bits in the flags field in the header.
enum {
	kZigZagRowsFlag = 1
};

// This is the structure of a RCI header:
// marker/magic number
// version number
// layout mode(interleavedrow,etc.)
// color mode(rgb, etc)
// number of planes(first planes are always color planes, rest are alphas)
// plane depth(1 isn't support, 8 for 24 bit RGB or 8 bit grayscale or 32 bit CMYK, 16, 1, etc)
// image size
// tile size(just set to zero if not applicable)
// filter selected(compression prefilter used)
// the rest of the file is ordered just like you would expect, there
// are no pad bytes whatsoever, so that we can experiment with
// small tile sizes and similiar things without it sucking cock.



typedef struct {
	// RciUns32 marker we don't store this, it isn't going to change
	RciUns32 dataStart; // start of image data relative to beginning of file allows 
		// later versions to add to header without breaking prior versions.
	RciUns16 version; // version of format
	RciUns16 planes; // number of planes
	RciUns8 colorMode; // color space
	RciUns8 planeDepth; // depth of each plane
	RciUns8 reserved; // pad value
	RciUns8 flags; // flags
	RciUns32 width, height; // image size
	RciUns32 tileWidth, tileHeight; // irrelevent if not tiled.
	
	// overall length of the header is 32 bytes.
	// This is the image data, in memory; image data in memory
	// is always interleaved, row major, and has no padding
	void *image;
	
	RciUns8 *tile; // This points into our scratch buffer while tiling image.

	BxBool filteringEnabled; // non-zero if filtering is enabled.
	int filter; // 0 for autoselect, specific filter otherwise.

	int fd; // file we are reading or writing from.

	// this is the compressed size of the image saved by RciReadImage and RciWriteImage
	// so that compression ratios can be calculated.
	RciUns32 compSize;
	// this is also set by RciWriteImageData and RciReadImageData
	// just for stats purposes.
	RciUns32 numTiles;

	// this is for gathering stats on 
	// the filter usage.	
	long filterStat[kRciNumFilters];

} RciState;


int RciReadImage(RciState *rci); // read a rci image 
int RciWriteImage(RciState *rci); // write out a rci image

int RciReadHeader(RciState *rci); // read in the header
int RciWriteHeader(RciState *rci); // write out the header
int RciReadImageData(RciState *rci); // alloc mem for data and read it
int RciWriteImageData(RciState *rci); // write data from memory

// Define this to the memcpy,memset,memcmp type routines of your choice.
#define RciMemCopy memcpy
#define RciMemSet memset
#define RciMemCmp memcmp

// These are defined away so you can plug into your own memory
// allocation functions if need be.
#define RciAlloc malloc
#define RciFree free


// everything within this ifdef is only needed by the library itself.
#ifdef RCI_INTERNAL

#if (defined(_WIN32) || defined(__WIN32__)) && !defined(WIN32)
#	define WIN32
#endif

// this has to be defined if on a little endian system and undef'ed on big endian system.
// Would be better if we autodetected more common configurations.
#ifdef WIN32
#define RCI_LITTLE_ENDIAN
#endif


// We just define this because sometimes nothing else does to keep code from breaking.
#ifndef NULL
#define NULL 0
#endif


// I/O functions
int RciRead8(int fd, RciUns8 *out);
int RciRead16BE(int fd, RciUns16 *out);
int RciRead32BE(int fd, RciUns32 *out);
int RciWrite8(int fd, RciUns8 in);
int RciWrite16BE(int fd, RciUns16 in);
int RciWrite32BE(int fd, RciUns32 in);

// these are the two functions you have to rewrite to port the I/O
int RciRead(int fd, RciUns8 *buf, RciUns32 len);
int RciWrite(int fd, const RciUns8 *buf, RciUns32 len);

// calculate the number of tiles per plane
RciUns32 RciNumTiles( RciState *rci);

// this reverses the odd rows in a tile
void RciZigZagTile(RciUns8 *tile, RciUns16 planeDepth, RciUns32 tileWidth, RciUns32 tileHeight);

// this writes a planar tile given its left,top and the tile width and height
// (it will clip the tile if need be)
int RciWritePlanarTile(RciState *rci, RciUns16 plane, RciUns32 left, RciUns32 top, RciUns32 *outBytes);

// this reads a planar tile given its left,top and the tile width and height
// (it will handled the clipped ones on right and bottom automatically)
int RciReadPlanarTile(RciState *rci, RciUns16 plane, RciUns32 left, RciUns32 top, RciUns32 *outBytes);

void RciCalcHist(RciUns8 *ptr, RciUns32 len, RciUns32 *hist);

RciUns8 RciFindMedian(RciUns8 *ptr, RciUns32 len);

// this filters a tile and returns the filter type used
RciUns8 RciFilterTile( int filter, RciUns8 *tile, RciUns16 planeDepth, RciUns32 rows, RciUns32 cols,
	RciUns8 *outFilterParam);

// this reverses the filter done on a tile.
void RciInvFilterTile( RciUns8 *tile, RciUns16 planeDepth, RciUns32 rows, RciUns32 cols,
	RciUns8 filterType, RciUns8 filterParam );


// this is the increase in size of the worst case scenario of the compression.
#define RCI_COMPRESS_MAX_INCREASE 4

RciInt32 RciDecompress(RciUns8 *input, RciUns32 insize, 
	RciUns8 *output, RciUns32 outsize );

RciInt32 RciCompress(RciUns8 *input,RciUns32 insize,
	RciUns8 *output,RciUns32 outsize);


#endif // RCI_INTERNAL

#ifdef __cplusplus
}
#endif


#endif
