/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define RCI_INTERNAL
#ifndef __RCI__
#include "rci.h"
#endif

#define MATCH_BITS 3
#define MIN_MATCH 3
#define LEN_ADD 2

#define BASE_LEN (((1<<MATCH_BITS)-1)+LEN_ADD)

#define MAX_MATCH (BASE_LEN+255)

#define LEN_MASK 0xE0 // 3 high bits of byte.

#define LIT_BITS 5
#define MAX_LIT (1<<LIT_BITS)
#define LIT_MASK (MAX_LIT-1)

#define HASH_BITS 16

#define HASH_SIZE (1L<<HASH_BITS)
#define HASH_MASK (HASH_SIZE-1)

// 16 bits
 #define HASH(a,b,c)  ( (a<<8) ^ (b<<4) ^ c )

#define OFFSET_BITS 13
#define WINDOW_SIZE (1L<<OFFSET_BITS)


// decompress it, outsize is how much output buffer can hold.
RciInt32 RciDecompress(
	RciUns8 *input, RciUns32 insize, 
	RciUns8 *output, RciUns32 outsize )
{
	RciUns32 decompsize;
	RciUns8 *inend = input+insize;
	RciUns8 *outstart = output, *outend = output+outsize;
	int bval;
	register int len;
	int offset;
	RciUns8 *matchptr;

	// decode size
	decompsize = *input++;
	decompsize <<= 8;
	decompsize |= *input++;
	decompsize <<= 8;
	decompsize |= *input++;
	decompsize <<= 8;
	decompsize |= *input++;

	if( decompsize & 0x80000000 )
	{
		// block is not compressed.
		decompsize &= 0x7FFFFFFF;
		
		if( decompsize > outsize )
			goto toobig;

		RciMemCopy(output,input,decompsize);
		return decompsize;
	}
	// check against outsize

	while( input < inend )
	{
		bval = *input++;
		if( bval & LEN_MASK )
		{
			// code.
			len = bval >> LIT_BITS;
			offset = (bval & LIT_MASK) << 8;
			
			if( input >= inend )
				goto compfail;
			offset |= *input++;
			
			len += LEN_ADD;

			matchptr = output - offset;
			if( matchptr < outstart )
				goto compfail;

			if( len == BASE_LEN )
			{
				if( input >= inend )
					goto compfail;
				len += *input++;
			}
			
			if( output+len > outend )
				goto toobig;

			while(len--)
				*output++ = *matchptr++;
		}
		else
		{
			// literals
			len = ++bval;
			if( output+len > outend )
				goto toobig;

			if( len == MAX_LIT )
			{
				do {
					if( input >= inend )
						goto compfail;
					len += (bval = *input++);
				}
				while( bval == 255 );
			}

			// Could probably make
			// this faster in most cases with an inline
			// copy even if it isn't as fast as the system
			// one that must use MMX instructions.
			RciMemCopy(output,input,len);
			input+=len;
			output+=len;
		}
	}

	// check that length matches.
	if( (RciUns32)(output-outstart) != decompsize )
		goto compfail;


	return decompsize;
toobig:
	// not enough output space.
	return -1;
compfail:
	// data corrupt
	return -1;
}

// We have to use inline on gcc, _inline in Visual C
#ifdef WIN32
_inline
#else
inline
#endif
RciUns8 *output_literals(RciUns8 *lit, RciUns8 *input, 
	RciUns8 *output, RciUns8 *outend)
{
	register RciInt32 inc, litstep, litcount = (RciInt32)(input - lit);

	litstep = litcount;
	if( litstep >= MAX_LIT )
	{
		if( output >= outend )
			return NULL;
		
		*output++ = MAX_LIT - 1;
		litstep -= MAX_LIT;
		do
		{
			if( output >= outend )
				return NULL;
		
			inc = litstep > 255 ? 255 : litstep;

			*output++ = (RciUns8)inc;
				
			litstep -= inc;
		}
		while( inc == 255 );
	}
	else
	{
		if( output >= outend )
			return NULL;

		*output++ = (RciUns8)(litstep - 1);	
	}
	if( output + litcount > outend )
		return NULL;

	while(litcount--)
		*output++ = *lit++;
	
	return output;
}

RciInt32 RciCompressCase2(
	RciUns8 *input,RciUns32 insize,
	RciUns8 *output,RciUns32 outsize)
{
	register int hval;
	register RciInt32 matchlen;
	RciInt32 offset;
	RciUns8 *hash[HASH_SIZE];
	RciUns8 *instart = input, *inend = input+insize, *instop;
	RciUns8 *outstart=output, *outend = output+outsize;
	register RciUns8 *cmp, *stop, *matchptr;
	RciUns8 *lit;

	if( insize < 3 )
		goto toobig;

	if( outsize < 7 )
		goto toobig;
	
	// put 4 byte decompressed size at beginning with high bit flag off
	*output++ = (RciUns8)( (insize>>24)&0xff );
	*output++ = (RciUns8)( (insize>>16)&0xff );
	*output++ = (RciUns8)( (insize>>8)&0xff );
	*output++ = (RciUns8)( insize&0xff );
	
	RciMemSet(hash,0,sizeof(RciUns8 *)*HASH_SIZE);

	lit = input;

	instop = inend - MAX_MATCH;
	
	while( input < instop )
	{
		// get hash key.
		hval = HASH(input[0],input[1],input[2]);

		// check for match, in range.
		if( //(hash[hval]) && 
			(input-hash[hval]) < WINDOW_SIZE &&
			hash[hval] && // ATTN: This line is what is different from RciCompressCase
#ifdef RCI_LITTLE_ENDIAN
			// the physical first 3 bytes are the little bytes
			(!(     ((*(RciUns32 *)hash[hval]) ^ (*(RciUns32 *)input)) & 0xFFFFFF))
#else
			// the physical first 3 bytes are the big bytes
			(!(     ((*(RciUns32 *)hash[hval]) ^ (*(RciUns32 *)input)) & 0xFFFFFF00))
#endif
			)
		{
			cmp = input;
			stop = input + MAX_MATCH;
			matchptr = hash[hval] + 3;
			offset=(RciInt32)(input-hash[hval]);

			// while also moving up input ptr.
			hash[hval] = cmp++;
			hash[HASH(cmp[0],cmp[1],cmp[2])] = cmp++;
			hash[HASH(cmp[0],cmp[1],cmp[2])] = cmp++;

			while( cmp < stop && *cmp == *matchptr )
			{
				hash[HASH(cmp[0],cmp[1],cmp[2])] = cmp++;
				matchptr++;
			}


			matchlen = (RciInt32)(cmp - input);
			{				
				// output any buffered literals.
				if( input != lit )
				{
					output = output_literals(lit,input,output,outend);
					if( !output )
						goto toobig;
				}

				// then output a match.
				// May only use 2 bytes, but one check is faster.
				if( output+3 > outend )
					goto toobig;

				if( matchlen >= BASE_LEN )
				{
					// we need to write a 3 byte code

					// low 5 of byte are high 5 of offset
					// high 3 of byte are 7
					*output++ = (RciUns8)( LEN_MASK | (offset >> 8) );
					// low 8 bits of offset
					*output++ = (RciUns8)( offset & 0xFF );
					// this byte is additional match length
					*output++ = (RciUns8)(matchlen - BASE_LEN);
				}
				else
				{
					// we can just write a 2 byte code.
					// high 3 bits are length
					// low 5 bits are high 5 bits of offset.
					*output++ = (RciUns8)( ((matchlen - LEN_ADD)<<5) | (offset >> 8));
					// low 8 bits of offset
					*output++ = (RciUns8)(offset & 0xFF);
				}

				lit = input = cmp;
				continue; // go to next time through.
			}
			// no match, fall through to literal case.
		}
		
		// update hash.
		hash[hval] = input;
		input++; // move up input ptr.
	} 

	// output remainder as literals.
	if( lit != inend )
	{
		output = output_literals(lit,inend,output,outend);
		if( !output )
			goto toobig;
	}

	return (RciInt32)(output-outstart);
toobig:
	// store size and memcopy it.
	if( outsize < 4+insize )
		return -1;

	output = outstart;

	// set high bit flag to indicate it's uncompressed.
	*output++ = (RciUns8)(((insize>>24)&0x7f) | 0x80);
	*output++ = (RciUns8)((insize>>16)&0xff);
	*output++ = (RciUns8)((insize>>8)&0xff);
	*output++ = (RciUns8)(insize&0xff);
	
	RciMemCopy(output,instart,insize);
	
	return insize+4;
}


RciInt32 RciCompress(
	RciUns8 *input,RciUns32 insize,
	RciUns8 *output,RciUns32 outsize)
{
	register int hval;
	register RciInt32 matchlen;
	RciInt32 offset;
	RciUns8 *hash[HASH_SIZE];
	RciUns8 *instart = input, *inend = input+insize, *instop;
	RciUns8 *outstart=output, *outend = output+outsize;
	register RciUns8 *cmp, *stop, *matchptr;
	RciUns8 *lit;

	// on most systems, 'input' will not located at a much higher virtual
	// address which allows us to drop one fork in the main loop
	// if for some reason, 'input' is really low then we go to the slower method
	if(input < (RciUns8 *)WINDOW_SIZE)
		return RciCompressCase2(input, insize, output, outsize);

	if( insize < 3 )
		goto toobig;

	if( outsize < 7 )
		goto toobig;
	
	// put 4 byte decompressed size at beginning
	// with high bit flag off
	*output++ = (RciUns8)((insize>>24)&0xff);
	*output++ = (RciUns8)((insize>>16)&0xff);
	*output++ = (RciUns8)((insize>>8)&0xff);
	*output++ = (RciUns8)(insize&0xff);

	RciMemSet(hash,0,sizeof(RciUns8 *)*HASH_SIZE);

	lit = input;
	instop = inend - MAX_MATCH;
	while( input < instop )
	{
		// get hash key.
		hval = HASH(input[0],input[1],input[2]);

		// check for match, in range.
		if( //(hash[hval]) && 
			(input-hash[hval]) < WINDOW_SIZE &&
#ifdef RCI_LITTLE_ENDIAN
			// the physical first 3 bytes are the little bytes
			(!(     ((*(RciUns32 *)hash[hval]) ^ (*(RciUns32 *)input)) & 0xFFFFFF))
#else
			// the phisical first 3 bytes are the big bytes
			(!(     ((*(RciUns32 *)hash[hval]) ^ (*(RciUns32 *)input)) & 0xFFFFFF00))
#endif
			)
		{
			cmp = input;
			stop = input + MAX_MATCH;
			matchptr = hash[hval] + 3;
			offset=(RciInt32)(input-hash[hval]);

			// while also moving up input ptr.
			hash[hval] = cmp++;
			hash[HASH(cmp[0],cmp[1],cmp[2])] = cmp++;
			hash[HASH(cmp[0],cmp[1],cmp[2])] = cmp++;

			// I think this might be slightly faster, but it's very arguable.
			while( cmp < stop && *cmp == *matchptr )
			{
				hash[HASH(cmp[0],cmp[1],cmp[2])] = cmp++;
				matchptr++;
			}


			matchlen = (RciInt32)(cmp - input);
			{
				
				// output any buffered literals.
				if( input != lit )
				{
					output = output_literals(lit,input,output,outend);
					if( !output )
						goto toobig;
				}

				// then output a match.


				// sometimes this only uses 2 bytes, but one check is faster.
				if( output+3 > outend )
					goto toobig;

				if( matchlen >= BASE_LEN )
				{
					// we need to write a 3 byte code

					// low 5 of byte are high 5 of offset
					// high 3 of byte are 7
					*output++ = (RciUns8)( LEN_MASK | (offset >> 8) );
					// low 8 bits of offset
					*output++ = (RciUns8)( offset & 0xFF );
					// this byte is additional match length
					*output++ = (RciUns8)(matchlen - BASE_LEN);
				}
				else
				{
					// we can just write a 2 byte code.
					// high 3 bits are length
					// low 5 bits are high 5 bits of offset.
					*output++ = (RciUns8)( ((matchlen - LEN_ADD)<<5) | (offset >> 8));
					// low 8 bits of offset
					*output++ = (RciUns8)(offset & 0xFF);
				}

				lit = input = cmp;

				continue; // go to next time through.
			}
			// no match, fall through to literal case.
		}
		// update hash.
		hash[hval] = input;
		input++; // move up input ptr.
	} 

	// output remainder as literals.
	if( lit != inend )
	{
		output = output_literals(lit,inend,output,outend);
		if( !output )
			goto toobig;
	}

	return (RciInt32)(output-outstart);
toobig:
	// store size and memcopy it.
	if( outsize < 4+insize )
		return -1;

	output = outstart;

	// set high bit flag to indicate it's uncompressed.
	*output++ = (RciUns8)(((insize>>24)&0x7f) | 0x80);
	*output++ = (RciUns8)((insize>>16)&0xff);
	*output++ = (RciUns8)((insize>>8)&0xff);
	*output++ = (RciUns8)(insize&0xff);
	
	RciMemCopy(output,instart,insize);	
	return insize+4;
}




