/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __RCI__
#include "rci.h"
#endif

// This test code is a little bit windows specific, but it shouldn't
// be hard to change for those on something else.
#include <windows.h> // only for timing.

#include <io.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

// test options:
// rci [-zsr] [wxh] billsit.rrgb billsit.rci
//	where w and h in wxh are width and height of tiles 
// rrgb billsit.rci billcopy.rrgb
// raw billsit.rci billcopy.raw
// To read in a raw input file, specify the filename and size together,
// like so: image.raw:100x80x3 (width x height x planes)

// test commands: 
// rci billsit.raw:427x641x3 billsit.rci
// rci billfive.raw:427x641x5 billfive.rci
// rci billgray.raw:427x641x1 billgray.rci

// non-zero indicates an error condition.
int ReadImage(RciState *rci); // read in an image of a supported format
int ReadRRGBImage(RciState *rci); // read an RRGB image
int WriteRRGBImage(RciState *rci); // write out an RRGB image

int main(int argc, char **argv)
{
	char *inname, *outname;
	char *tokptr;
	char *fmtArg;
	int opt;
	int optind;
	char *argp;
	int ret, err = 0;
	int fmt = kRciNativeFmt;
	int tileWidth = kRciDefaultTileWidth;
	int tileHeight = kRciDefaultTileHeight;
	RciState rci;
	RciUns8 flags = 0;
	int filter = 0; // autoselect.
	int rawWidth=0,rawHeight=0, rawPlanes=0;
	char *ss;

	// these vars are for timing and compression stats.
	RciUns32 fileLen;
	double dtemp;
	long ltemp;
	DWORD bytes, startTicks, stopTicks, totalTicks=0;
	int plane;

	
	// set things up in a state where we can tell what we need to cleanup and RciFree.
	rci.image = NULL;
	rci.tile = NULL;
	rci.fd = -1;
	

	rci.filteringEnabled = 0;

	// output format - rrgb or rci or raw
	// flags we need to handle: 
	// WxH - width by height of tiles
	// z - zig zag rows if present
	// s yes/no - filtering or not.
	// r - force tiles to be entire single rows
	// f# - where # is filter to use by #, 0 is autoselect.

	// so usage will be: rrgb infile outfile || rci -[zsr] [WxH] infile outfile
	if( argc < 4 )
		goto usage;

	optind = 1; // skip the program name.
	fmtArg = argv[optind++];
	if( !strcmp(fmtArg,"rrgb") )
		fmt = kRciRRGBFmt;
	else if( !strcmp(fmtArg,"rci") )
		fmt = kRciNativeFmt;
	else if( !strcmp(fmtArg,"raw") )
		fmt = kRciRawFmt;
	else
	{
		fprintf(stderr,"Not a known output format\n");
		goto usage;
	}

	
	if( fmt == kRciNativeFmt )
	{
		// check for optional args.
		if( argc - optind > 2 )
		{
			if( argv[optind][0] == '-' )
			{
				argp = argv[optind++];
				// this is the flags argument
				argp++;
				while( (opt=*argp++) != 0 )
				{
					switch(opt)
					{
						case 'z':
							// zig zag the rows.
							flags |= kZigZagRowsFlag;
							break;
						case 's': // filtering option, relavent chars were taken
							rci.filteringEnabled = 1;
							break;
						case 'r': // force tile sizes to 1 whole row each
							tileHeight = 1;
							tileWidth = -1;
							break;
						case 'f':
							filter = *argp++;
							if( !filter || filter < '0' || filter >= ('0' + kRciNumFilters) )
								goto usage;
							filter -= '0'; // from ascii

							break;
						default:
							goto usage; // bad option
					}					
				}
			} // if( argp[0] == '-' )
		} // if( argc - optind )
		if( argc - optind > 2)
		{
			// this is the tilesize argument.
			argp = argv[optind++];
			tokptr = strtok( argp, "x" );
			tileWidth = atoi(argp);
			tokptr = strtok( NULL, "x" );
			if( !tokptr )
				tileHeight = tileWidth;
			else
				tileHeight = atoi( tokptr );
			
		}
	}
	else if( fmt == kRciRRGBFmt || fmt == kRciRawFmt )
	{
		if( argc - optind != 2 ) // should be no args left.
			goto usage;
	}

	inname = argv[optind++];
	outname = argv[optind++];

	printf("%s ---> %s\n",inname,outname);

	// handle size specification for raw files.
	if( (ss=strstr(inname,".raw:")) != 0 )
	{
		ss[4] = 0; // terminate after actual name.
		ss += 5; // skip to width x height

		tokptr = strtok( ss, "x" );
		rawWidth = atoi(ss);
		tokptr = strtok( NULL, "x" );
		if( !tokptr )
		{
			rawHeight = rawWidth;
			rawPlanes = 3;
		}
		else
		{
			rawHeight = atoi( tokptr );
			// get planes.
			tokptr = strtok( NULL, "x" );
			rawPlanes = tokptr ? atoi(tokptr) : 3;
		}

	}
	// read the input file.
	ret = _open(inname, _O_RDONLY | _O_BINARY );
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	rci.fd = ret;

	if( rawWidth && rawHeight && rawPlanes )
	{
		rci.image = RciAlloc(rawWidth*rawHeight*rawPlanes);
		if( !rci.image )
		{
			err = kRciNoMem;
			goto cleanup;
		}
		ret = _read(rci.fd,rci.image,rawWidth*rawHeight*rawPlanes);
		if( ret == -1 )
		{
			err = kRciReadErr;
			goto cleanup;
		}
		else if( ret != rawWidth*rawHeight*rawPlanes)
		{
			err = kRciEOFErr;
			goto cleanup;
		}
	
		rci.width = rawWidth;
		rci.height = rawHeight;
		rci.planes = rawPlanes;
		// assume RGB if 3 or more planes.
		rci.colorMode = ( rawPlanes >= 3 ) ? kRciRGB : kRciGrayscale;
		rci.planeDepth = 8; // only support 8 bit right now.
		// that should do it.
	}
	else
	{
		err = ReadImage( &rci );
		if( err )
			goto cleanup;
	}

	ret = _close(rci.fd);
	rci.fd = -1;
	if( ret )
	{
		err = errno;
		goto cleanup;
	}

	// set the output mode.
	// rci.fileMode = mode;
	// if we are tiling it, and tile size wasn't set in read file,
	// default it in output.

	rci.filter = filter;

	if( tileWidth == -1 )
		rci.tileWidth = rci.width; // special case to do row sized tiles.
	else
		rci.tileWidth = tileWidth;//Size;
	rci.tileHeight = tileHeight;//Size;
	rci.flags = flags;

	printf("Tile size %dx%d\n",rci.tileWidth,rci.tileHeight);

	// write the output file.
	ret = open(outname,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0666);
	if( ret == -1 )
	{
		err = errno;
		goto cleanup;
	}
	rci.fd = ret;
	if( fmt == kRciNativeFmt )
	{

		// Time it.
		startTicks = GetTickCount();
	
		// write the image.
		err = RciWriteImage(&rci);

		stopTicks = GetTickCount();
		totalTicks = stopTicks - startTicks;

		bytes = rci.planes * (rci.planeDepth>>3) * rci.width * rci.height;
		printf("%u bytes compressed in %u ticks\n",bytes,totalTicks);

		// print out stats on compression ratio.
		fileLen = rci.compSize;
		
		dtemp = fileLen * 100;
		dtemp /= bytes;
		printf("%g%% of raw image size\n",dtemp);

		// print out the stats for filter usage.
		ltemp = rci.numTiles * 3;
		// if autoselecting filter, print out some stats.
		if( rci.filteringEnabled && rci.filter == 0 )
		{	
			for(plane=0;plane<kRciNumFilters;plane++) // just using plane cause it's convenient
			{
				dtemp = rci.filterStat[plane] * 100;
				dtemp /= ltemp;
				printf("%g%% tiles with filter type %d\n",dtemp,(int)plane );
			}
		}


	}
	else if( fmt == kRciRRGBFmt )
		err = WriteRRGBImage(&rci);
	else if( fmt == kRciRawFmt )
	{
		if( _write(rci.fd,(void *)rci.image,rci.width*rci.height*rci.planes) != rci.width*rci.height*rci.planes )
			err = kRciWriteErr;
	}
	else
		err = kRciStateErr;
	
	if( err )
		goto cleanup;
	
	ret = _close(rci.fd);
	rci.fd = -1;
	if( ret )
	{
		err = errno;
		goto cleanup;
	}

cleanup:
	if( err > 0 )
		fprintf(stderr,"%s,%d\n",strerror(err),err);
	else if( err < 0 )
	{
		switch(err)
		{
			case kRciNotFmtErr:
				fprintf(stderr,"%s doesn't appear to be in a supported format\n",inname);
				break;
			case kRciEOFErr:
				fprintf(stderr,"Unexpected end-of-file error\n");
				break;
			default:
				fprintf(stderr,"Unknown error %d\n",err);
		}
	}

	if( rci.fd != -1 )
		(void)_close(rci.fd);
	if( rci.image )
		RciFree( rci.image );
	
	return err ? 1 : 0;
usage:
	// Would be nice if we actually told them the usage.
	fprintf(stderr,"See usage!\n");
	return 2;
}


int ReadImage(RciState *rci) // read in an image of a supported format
{
	DWORD bytes, startTicks, stopTicks, totalTicks=0;
	// we have written our reading functions in such a way
	// that we can just keep calling them until they stop returning
	// kRciNotFmtErr, provided we seek back to the start of the
	// file in between tries.

	int err = 0;
	err = ReadRRGBImage(rci);
	// success, or actual real error, return.
	if( err != kRciNotFmtErr )
		return err;
	
	// seek back to start
	lseek(rci->fd,0,SEEK_SET);


	startTicks = GetTickCount();
	err = RciReadImage(rci);
	if( err )
		return err;

	stopTicks = GetTickCount();
	totalTicks = stopTicks - startTicks;

	bytes = rci->planes * (rci->planeDepth>>3) * rci->width * rci->height;
	printf("%u bytes decompressed in %u ticks\n",bytes,totalTicks);

	return err;
}


